{ Description:
  Security Services user id, password, and simple access data component.
  Provides simple UserId/Password valdiation and registration using the
  Windows registry as persistent storage (database).

  Author:
  Thomas Hill, dba t.h.ink Software
  hillt@ionet.net

  History:
  01/21/96 - Version 0.1.  First working version.
             Basic methods and properties completed.
             Help file for existing properties, methods, and types completed.
             Test of properties and methods passsed.
             Event handlers tests passed.
  01/23/96 - Version 1.0  Final release version.
             Added DeleteUserID method (Forgot this!)
             Added DeleteAccessData method (Forgot this one, too!)
             Expanded ESecurityError class structure, to better define possible
             error conditions to the user's program.
             This completes the development for this component.  Further
             functionality will be encapsulated within other Security Components.

  Comments:
  This is a simplistic starting point for the full Security Server I want to
  develop.  It is, even in this incarnation, a useful component, so I am
  releasing it to the world.  Further components along these lines will be
  forthcoming, as I complet and finalize the design document for the Security
  Server.  Components like this one are points on the learning curve to the
  final product.  
}

unit Secbase;

interface

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
  Forms, Dialogs, ShellAPI;

type
  ESecurityError = class(Exception);
  ESecurityBadApp = class(ESecurityError);
  ESecurityBadUser = class(ESecurityError);
  ESecurityBadPW = class(ESecurityError);
  ESecurityBadDB = class(ESecurityError);

  TEncryptEvent = procedure(PlainText : string; var CipherText : string) of object;
  TDecryptEvent = procedure(CipherText : string; var PlainText : string) of object;
  TThSecBase = class(TComponent)
  private
    FActive : boolean;
    FAppKey : hKey;
    FAppName : string;
    FRootKey : hKey;
    FUseCase : boolean;
    FOnEncrypt : TEncryptEvent;
    FOnDecrypt : TDecryptEvent;
    procedure InternalEncrypt(PlainText : string; var CipherText : string);
    procedure InternalDecrypt(CipherText : string; var PlainText : string);

  protected
    procedure SetAppName(Value : string);

  public
    constructor Create(AOwner : TComponent); override;
    destructor Destroy; override;
    procedure DeleteAccessData(UserID, Password : string; var Success : boolean);
    procedure DeleteUserID(UserID, Password : string; var Success : boolean);
    procedure RegisterUserID(UserID,Password : string);
    procedure RegisterAccessData(UserID,Password,Data : string);
    procedure ValidateUserID(UserID, Password : string; var Valid : boolean);
    procedure GetAccessData(UserId,Password : string; var Data : string);

  published
    property Active : boolean read FActive;
    property ApplicationName : string read FAppName write SetAppName;
    property CaseSensitive : boolean read FUseCase write FUseCase;
    property OnEncrypt : TEncryptEvent read FOnEncrypt write FOnEncrypt;
    property OnDecrypt : TDeCryptEvent read FOnDecrypt write FOnDecrypt;

  end;

procedure Register;

implementation

const
   Norm : string = '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz';
   Rot  : string = 'NOPQRSTUVWXYZ0123456789ABCDEFGHIJKLMnopqrstuvwxyzabcdefghijklm';
   RegSecRoot : pChar = 'Software\t.h.ink\SecServer';
{ These were lifted from the WINREG.H file from BC++ 4.52 }
   HKEY_CLASSES_ROOT  = $80000000;
   HKEY_CURRENT_USER  = $80000001;
   HKEY_LOCAL_MACHINE = $80000002;
   HKEY_USERS         = $80000003;


procedure Register;
begin
  RegisterComponents('Custom', [TThSecBase]);
end;

{ Internal encrpyt/decrypt uses the Usenet ROT13 method.  This is child's play,
  and should DEFINTELY be replaced with something a little bit more secure!
}
procedure TThSecBase.InternalEncrypt(PlainText : string; var CipherText : string);

var
   WorkText : string;
   i,j : integer;
   Found : boolean;

begin
  for i := 1 to length(PlainText) do
  begin
    Found := FALSE; j := 1;
    repeat
      if Norm[j] = PlainText[i] then Found := TRUE;
      inc(j);
    until Found or (j > 62);
    if Found then WorkText[i] := Rot[j - 1] else WorkText[i] := PlainText[i];
  end;
  WorkText[0] := char(length(PlainText));
  CipherText := WorkText;
end;

{ Internal encrpyt/decrypt uses the Usenet ROT13 method.  This is child's play,
  and should DEFINTELY be replaced with something a little bit more secure!
}
procedure TThSecBase.InternalDecrypt(CipherText : string; var PlainText : string);

var
   WorkText : string;
   i,j : integer;
   Found : boolean;

begin
  for i := 1 to length(CipherText) do
  begin
    Found := FALSE; j := 1;
    repeat
      if Rot[j] = CipherText[i] then Found := TRUE;
      inc(j);
    until Found or (j > 62);
    if Found then WorkText[i] := Norm[j - 1] else WorkText[i] := CipherText[i];
  end;
  WorkText[0] := char(length(CipherText));
  PlainText := WorkText;
end;

constructor TThSecBase.Create(AOwner : TCOmponent);

begin
  inherited Create(AOwner);
  FRootKey := 0;
  FAppName := '';
  FActive := FALSE;
  if RegCreateKey(HKEY_CURRENT_USER,RegSecRoot,FRootKey) <> ERROR_SUCCESS then
  begin
    raise ESecurityBadDB.Create('Unable to access Security Registry Root!');
  end;
end;

destructor TThSecBase.Destroy;

var
   szStr : pChar;

begin
  getmem(szStr,8);
  if FAppKey <> 0 then
  begin
    StrCopy(szStr,'Closed');
    RegSetValue(FAppKey,nil,REG_SZ,szStr,strlen(szStr));
    RegCloseKey(FAppKey);
  end;
  if FRootKey <> 0 then RegCloseKey(FRootKey);
  freemem(szStr,8);
  inherited Destroy;
end;

procedure TThSecBase.SetAppName(Value : string);

var
   szStr : pChar;

begin
  try
    getmem(szStr,255);
    FActive := FALSE;
    if FRootKey <> 0 then
    begin
      if length(FAppName) > 0 then
      begin
        StrCopy(szStr,'Closed');
        RegSetValue(FAppKey,nil,REG_SZ,szStr,strlen(szStr));
        FActive := FALSE;
        RegCloseKey(FAppKey);
      end;
      FAppName := Value;
      StrPCopy(szStr,FAppName);
      if RegCreateKey(FRootKey,szStr,FAppKey) <> ERROR_SUCCESS then
      begin
        FAppName := '';
        raise ESecurityBadApp.CreateFmt('Could not create/open Registry Entry for %s',[FAppName]);
      end
      else
      begin
        StrCopy(szStr,'Open');
        RegSetValue(FAppKey,nil,REG_SZ,szStr,strlen(szStr));
        FActive := TRUE;
      end;
    end
    else
      raise ESecurityBadDB.Create('Security Registry Root key invalid!');
  finally
    freemem(szStr,255);
  end;
end;

procedure TThSecBase.RegisterUserID(UserID, Password : string);

var
   szKey,szValue : pChar;
   CryptText : string;

begin
  try
    getmem(szKey,255); getmem(szValue,255);
    if (FRootKey <> 0) and (FAppKey <> 0) then
    begin
      StrPCopy(szKey,UserId); StrCat(szKey,'\Password');
      if assigned(FOnEncrypt) then
        FOnEncrypt(Password,CryptText)
      else
        InternalEncrypt(Password,CryptText);
      StrPCopy(szValue,CryptText);
      if RegSetValue(FAppKey,szKey,REG_SZ,szValue,strlen(szValue)) <> ERROR_SUCCESS then
        raise ESecurityBadUser.CreateFmt('Could not register userID %s',[UserID]);
    end
    else
      raise ESecurityBadDB.Create('Security Root Key or Applicaiton Key invalid!');
  finally
    freemem(szKey,255); freemem(szValue,255);
  end;
end;

procedure TThSecBase.ValidateUserID(UserId,Password : string; var Valid : boolean);

var
   szKey,szValue,szTemp : pChar;
   cbSize : longint;
   CText,PText : string;

begin
  try
    Valid := FALSE;
    getmem(szKey,255); getmem(szValue,255); getmem(szTemp,255);
    if FActive then
    begin
      StrPCopy(szKey,UserId); StrCat(szKey,'\Password');
      fillchar(szValue,0,255);
      cbSize := 255;
      if RegQueryValue(FAppKey,szKey,szValue,cbSize) <> ERROR_SUCCESS then
        raise ESecurityBadUser.CreateFmt('Could not retrieve data for UserID %s',[UserID])
      else
      begin
        CText := StrPas(szValue);
        if Assigned(FOnDecrypt) then
          FOnDecrypt(CText,PText)
        else
          InternalDecrypt(CText,PText);
        if FUseCase then
          Valid := (CompareStr(PText,Password) = 0)
        else
          Valid := (CompareText(PText,Password) = 0);
      end;
    end
    else
      raise ESecurityBadApp.Create('No active application');
  finally
    freemem(szKey,255); freemem(szValue,255); freemem(szTemp,255);
  end;
end;

procedure TThSecBase.DeleteUserID(UserID, Password : string; var Success : boolean);

var
   szKey : pChar;
   PWOk : boolean;

begin
  try
    getmem(szKey,255);
    if FActive then
    begin
      ValidateUserID(UserID,Password,PWOk);
      if PWOk then
      begin
        StrPCopy(szKey,UserId);
        if RegDeleteKey(FAppKey,szKey) = ERROR_SUCCESS then
        begin
          Success := TRUE
        end
        else
        begin
          Success := FALSE;
          raise ESecurityBadDB.CreateFmt('Could not delete User ID %s',[UserID])
        end;
      end
      else
        raise ESecurityBadPW.Create('Access denied.');
    end
    else
      raise ESecurityBadDB.Create('No active aplication');
  finally
    freemem(szKey,255);
  end;
end;


procedure TThSecBase.RegisterAccessData(UserID,Password,Data : string);

var
   szKey,szValue : pChar;
   PWOk : boolean;
   CryptText : string;

begin
  try
    getmem(szKey,255); getmem(szValue,255);
    if (FRootKey <> 0) and (FAppKey <> 0) then
    begin
      ValidateUserID(UserID,Password,PWOk);
      if PWOk then
      begin
        StrPCopy(szKey,UserId); StrCat(szKey,'\Access');
        if assigned(FOnEncrypt) then
          FOnEncrypt(Data,CryptText)
        else
          InternalEncrypt(Data,CryptText);
        StrPCopy(szValue,CryptText);
        if RegSetValue(FAppKey,szKey,REG_SZ,szValue,strlen(szValue)) <> ERROR_SUCCESS then
          raise ESecurityBadDB.CreateFmt('Could not register Access Data for %s',[UserID]);
      end
      else
        raise ESecurityBadPW.Create('Access denied.');
    end
    else
      raise ESecurityBadDB.Create('Security Root Key or Applicaiton Key invalid!');
  finally
    freemem(szKey,255); freemem(szValue,255);
  end;
end;

procedure TThSecBase.GetAccessData(UserID,Password : string; var Data : string);

var
   szKey,szValue,szTemp : pChar;
   cbSize : longint;
   PWOk : boolean;
   CTExt,PText : string;

begin
  try
    getmem(szKey,255); getmem(szValue,255); getmem(szTemp,255);
    if FActive then
    begin
      ValidateUserID(UserID,Password,PWOk);
      if PWOk then
      begin
        StrPCopy(szKey,UserId); StrCat(szKey,'\Access');
        fillchar(szValue,0,255);
        cbSize := 255;
        if RegQueryValue(FAppKey,szKey,szValue,cbSize) = ERROR_SUCCESS then
        begin
          CText := StrPas(szValue);
          if Assigned(FOnDecrypt) then
            FonDecrypt(CText,PText)
          else
            InternalDecrypt(CText,PText);
          Data := PText;
        end
        else
          raise ESecurityBadDB.CreateFmt('Could not retrieve access data for UserID %s',[UserID])
      end
      else
        raise ESecurityBadPW.Create('Access denied.');
    end
    else
      raise ESecurityBadApp.Create('No active aplication');
  finally
    freemem(szKey,255); freemem(szValue,255); freemem(szTemp,255);
  end;
end;

procedure TThSecBase.DeleteAccessData(UserID, Password : string; var Success : boolean);

var
   szKey : pChar;
   PWOk : boolean;

begin
  try
    getmem(szKey,255);
    if FActive then
    begin
      ValidateUserID(UserID,Password,PWOk);
      if PWOk then
      begin
        StrPCopy(szKey,UserId); StrCat(szKey,'\Access');
        if RegDeleteKey(FAppKey,szKey) = ERROR_SUCCESS then
        begin
          Success := TRUE
        end
        else
        begin
          Success := FALSE;
          raise ESecurityBadDB.CreateFmt('Could not delete access data for UserID %s',[UserID])
        end;
      end
      else
        raise ESecurityBadPW.Create('Access denied.');
    end
    else
      raise ESecurityBadApp.Create('No active aplication');
  finally
    freemem(szKey,255);
  end;
end;


end.
