{
   Component: TThSecTime
   Description: Part of the Security Components developed by t.hink Software.
                This component provides time constraint validation for Users
                who use the TSecBase component to manage User IDs and Passwords.
                TThSecTime stores time constraint information in the Windows
                registery as pairs of times, specifying a begin and end
                time for each day of the week.  An internal timer may be activated
                that will trigger an OnWathTimer event which the user may assign
                a mthdo to to re-validate user time access authority.

   History : 01/24/96 - Version 0.1  Begin work.  Carry over much of TSecBase.
             01/27/96 - Version 0.2 First working version for palette
                        Implement RegisterUserTimes, ValidateUserTimes methods,
                        OnWatchTimer event, WatchTime property. Expand ROT13 to
                        handle digits.
             01/29/96 - Version 0.21
                        Added DeleteUserTimes method.  Why do I always forget the
                        delete functions??
}

unit Sectime;

interface

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
  Forms, Dialogs, ExtCtrls, ShellApi;

type
  ESecurityError = class(Exception);
  ESecurityBadApp = class(ESecurityError);
  ESecurityBadUser = class(ESecurityError);
  ESecurityBadPW = class(ESecurityError);
  ESecurityBadDB = class(ESecurityError);

  TEncryptEvent = procedure(PlainText : string; var CipherText : string) of object;
  TDecryptEvent = procedure(CipherText : string; var PlainText : string) of object;
  TThSecTime = class(TComponent)
  private
    FActive : boolean;
    FRootKey : hKey;
    FAppKey : hKey;
    FAppName : string;
    FUseCase : boolean;
    FStartTime : TDateTime;
    FEndTime : TDateTime;
    FWatching : boolean;  { flag to indicate we are watching the passage of time }
    FWatchInterval : integer;  { this is in MINUTES, not milliseconds }
    FWatchTimer : TTimer;
    FOnEncrypt : TEncryptEvent;
    FOnDecrypt : TDecryptEvent;
    FOnTimer : TNotifyEvent;
    procedure InternalEncrypt(PlainText : string; var CipherText : string);
    procedure InternalDecrypt(CipherText : string; var PlainText : string);
    procedure ValidateUserID(UserID, Password : string; var Valid : boolean);
    procedure WatchTimer(Sender : TObject);

  protected
    procedure SetAppName(Value : string);
    procedure SetWatchTimer(Value :boolean);
    procedure SetWatchInterval(Value : integer);

  public
    constructor Create(AOwner : TComponent); override;
    destructor Destroy; override;
    procedure RegisterUserTimes(UserID, Password : string; WeekDay : integer;
                                Start,Finish : TDateTime);
    procedure ValidateUserTimes(UserID, Password : string; var Valid : boolean);
    procedure DeleteUserTimes(UserID, Password : string; WeekDay : integer;
                              var Success : boolean);

  published
    property Active : boolean read FActive;
    property ApplicationName : string read FAppName write SetAppName;
    property CaseSensitive : boolean read FUseCase write FUseCase default FALSE;
    property WatchInterval : integer read FWatchInterval write SetWatchInterval default 1;
    property WatchTime : boolean read FWatching write SetWatchTimer default FALSE;
    property OnEncrypt : TEncryptEvent read FOnEncrypt write FOnEncrypt;
    property OnDecrypt : TDeCryptEvent read FOnDecrypt write FOnDecrypt;
    property OnWatchTimer : TNotifyEvent read FOnTimer write FOnTimer;

  end;

procedure Register;

implementation

const
   { This is used ONLY for Password verification.
     See also the OnEncrypt and OnDecrypt Events }
   Norm : string = '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz';
   Rot  : string = 'NOPQRSTUVWXYZ0123456789ABCDEFGHIJKLMnopqrstuvwxyzabcdefghijklm';
   WDays : array[1..7] of string =
           ('\Sunday','\Monday','\Tuesday','\Wednesday','\Thursday','\Friday','\Saturday');

   RegSecRoot : pChar = 'Software\t.h.ink\SecServer';
{ These were lifted from the WINREG.H file from BC++ 4.52 }
   HKEY_CLASSES_ROOT  = $80000000;
   HKEY_CURRENT_USER  = $80000001;
   HKEY_LOCAL_MACHINE = $80000002;
   HKEY_USERS         = $80000003;

var
   MinuteCnt : integer;

procedure Register;
begin
  RegisterComponents('Custom', [TThSecTime]);
end;

{ Internal encrpyt/decrypt uses the Usenet ROT13 method.  This is child's play,
  and should DEFINITELY be replaced with something a little bit more secure!
}
procedure TThSecTime.InternalEncrypt(PlainText : string; var CipherText : string);

var
   WorkText : string;
   i,j : integer;
   Found : boolean;

begin
  for i := 1 to length(PlainText) do
  begin
    Found := FALSE; j := 1;
    repeat
      if Norm[j] = PlainText[i] then Found := TRUE;
      inc(j);
    until Found or (j > 62);
    if Found then WorkText[i] := Rot[j - 1] else WorkText[i] := PlainText[i];
  end;
  WorkText[0] := char(length(PlainText));
  CipherText := WorkText;
end;

{ Internal encrpyt/decrypt uses the Usenet ROT13 method.  This is child's play,
  and should DEFINITELY be replaced with something a little bit more secure!
}
procedure TThSecTime.InternalDecrypt(CipherText : string; var PlainText : string);

var
   WorkText : string;
   i,j : integer;
   Found : boolean;

begin
  for i := 1 to length(CipherText) do
  begin
    Found := FALSE; j := 1;
    repeat
      if Rot[j] = CipherText[i] then Found := TRUE;
      inc(j);
    until Found or (j > 62);
    if Found then WorkText[i] := Norm[j - 1] else WorkText[i] := CipherText[i];
  end;
  WorkText[0] := char(length(CipherText));
  PlainText := WorkText;
end;

constructor TThSecTime.Create(AOwner : TCOmponent);

begin
  inherited Create(AOwner);
  FRootKey := 0;
  FAppName := '';
  FActive := FALSE;
  FWatchInterval := 1;  { every minute }
  if RegOpenKey(HKEY_CURRENT_USER,RegSecRoot,FRootKey) <> ERROR_SUCCESS then
  begin
    raise ESecurityBadDB.Create('Unable to access Security Registry Root!');
  end;
  FWatchTimer := TTimer.Create(Self);
  FWatchTimer.OnTimer := WatchTimer;
  FWatchTimer.Interval := 60000;  { timer triggers every minute }
  FWatchTimer.Enabled := FALSE;
  MinuteCnt := 0;
end;

destructor TThSecTime.Destroy;

var
   szStr : pChar;

begin
  getmem(szStr,8);
  if FAppKey <> 0 then
  begin
    StrCopy(szStr,'Closed');
    RegSetValue(FAppKey,nil,REG_SZ,szStr,strlen(szStr));
    RegCloseKey(FAppKey);
  end;
  if FRootKey <> 0 then RegCloseKey(FRootKey);
  freemem(szStr,8);
  FWatchTimer.Free;
  inherited Destroy;
end;

procedure TThSecTime.SetAppName(Value : string);

var
   szStr : pChar;

begin
  try
    getmem(szStr,255);
    FActive := FALSE;
    if FRootKey <> 0 then
    begin
      if length(FAppName) > 0 then
      begin
        StrCopy(szStr,'Closed');
        RegSetValue(FAppKey,nil,REG_SZ,szStr,strlen(szStr));
        FActive := FALSE;
        RegCloseKey(FAppKey);
      end;
      FAppName := Value;
      StrPCopy(szStr,FAppName);
      if RegOpenKey(FRootKey,szStr,FAppKey) <> ERROR_SUCCESS then
      begin
        FAppName := '';
        raise ESecurityBadApp.CreateFmt('Could not open Registry Entry for %s',[FAppName]);
      end
      else
      begin
        StrCopy(szStr,'Open');
        RegSetValue(FAppKey,nil,REG_SZ,szStr,strlen(szStr));
        FActive := TRUE;
      end;
    end
    else
      raise ESecurityBadDB.Create('Security Registry Root key invalid!');
  finally
    freemem(szStr,255);
  end;
end;

procedure TThSecTime.ValidateUserID(UserId,Password : string; var Valid : boolean);

var
   szKey,szValue,szTemp : pChar;
   cbSize : longint;
   CText,PText : string;

begin
  try
    Valid := FALSE;
    getmem(szKey,255); getmem(szValue,255); getmem(szTemp,255);
    if FActive then
    begin
      StrPCopy(szKey,UserId); StrCat(szKey,'\Password');
      fillchar(szValue,0,255);
      cbSize := 255;
      if RegQueryValue(FAppKey,szKey,szValue,cbSize) <> ERROR_SUCCESS then
        raise ESecurityBadUser.CreateFmt('Could not retrieve data for UserID %s',[UserID])
      else
      begin
        CText := StrPas(szValue);
        if Assigned(FOnDecrypt) then
          FOnDecrypt(CText,PText)
        else
          InternalDecrypt(CText,PText);
        if FUseCase then
          Valid := (CompareStr(PText,Password) = 0)
        else
          Valid := (CompareText(PText,Password) = 0);
      end;
    end
    else
      raise ESecurityBadApp.Create('No active application');
  finally
    freemem(szKey,255); freemem(szValue,255); freemem(szTemp,255);
  end;
end;

procedure TThSecTime.SetWatchTimer(Value : boolean);

begin
  if Value <> FWatching then
  begin
    FWatching := Value;
    FWatchTimer.Enabled := Value;
  end;
end;

procedure TThSecTime.RegisterUserTimes(UserID, Password : string; WeekDay : integer;
                                       Start,Finish : TDateTime);

var
   szKey,szValue : pChar;
   PWOk : boolean;
   PlainTime : string;
   CryptTime : string;

begin
  try
    getmem(szKey,255); getmem(szValue,255);
    if (FRootKey <> 0) and (FAppKey <> 0) then
    begin
      ValidateUserID(UserID,Password,PWOk);
      if PWOk then
      begin
        StrPCopy(szKey,UserId); StrCat(szKey,'\Time Constraints');
        StrPCopy(szValue,WDays[WeekDay]); StrCat(szKey,szValue);
        PlainTime := FormatDateTime('hhnn',Start);
        PlainTime := PlainTime + FormatDateTime('hhnn',Finish);
        if assigned(FOnEncrypt) then
          FOnEncrypt(PlainTime,CryptTime)
        else
          InternalEncrypt(PlainTime,CryptTime);
        StrPCopy(szValue,CryptTime);
        if RegSetValue(FAppKey,szKey,REG_SZ,szValue,strlen(szValue)) <> ERROR_SUCCESS then
          raise ESecurityBadDB.CreateFmt('Could not register Time Constraint for %s',[UserID]);
      end
      else
        raise ESecurityBadPW.Create('Access denied.');
    end
    else
      raise ESecurityBadDB.Create('Security Root Key or Application Key invalid!');
  finally
    freemem(szKey,255); freemem(szValue,255);
  end;
end;

procedure TThSecTime.ValidateUserTimes(UserID, Password : string; var Valid : boolean);

var
   szKey,szValue : pChar;
   PWOk : boolean;
   cbSize : longint;
   WDay : integer;
   PlainTime : string;
   CryptTime : string;
   Start,Finish : TDateTime;

begin
  try
    Valid := FALSE;
    getmem(szKey,255); getmem(szValue,255);
    if (FRootKey <> 0) and (FAppKey <> 0) then
    begin
      ValidateUserID(UserID,Password,PWOk);
      if PWOk then
      begin
        StrPCopy(szKey,UserId); StrCat(szKey,'\Time Constraints');
        WDay := DayOfWeek(Now);
        StrPCopy(szValue,WDays[WDay]); StrCat(szKey,szValue);
        fillchar(szValue^,0,255); cbSize := 255;
        if RegQueryValue(FAppKey,szKey,szValue,cbSize) <> ERROR_SUCCESS then
          raise ESecurityBadDB.CreateFmt('Could not retrieve Time Constraint for %s',[UserID]);
        CryptTime := StrPas(szValue);
        if assigned(FOnDecrypt) then
          FOnDecrypt(CryptTime,PlainTime)
        else
          InternalDecrypt(CryptTime,PlainTime);
        Start := StrToTime(copy(PlainTime,1,2) + ':' + copy(PlainTime,3,2));
        Finish := StrToTime(copy(PlainTime,5,2) + ':' + copy(PlainTime,7,2));
        if (Start <= Time) and (Time <= Finish) then Valid := TRUE;
      end
      else
        raise ESecurityBadPW.Create('Access denied.');
    end
    else
      raise ESecurityBadDB.Create('Security Root Key or Application Key invalid!');
  finally
    freemem(szKey,255); freemem(szValue,255);
  end;
end;

procedure TThSecTime.DeleteuserTimes(UserID, Password : string; WeekDay : integer;
                                     var Success : boolean);

var
   szKey,szTemp : pChar;
   PWOk : boolean;

begin
  try
    getmem(szKey,255);
    getmem(szTemp,255);
    Success := FALSE;
    if (FRootKey <> 0) and (FAppKey <> 0) then
    begin
      ValidateUserID(UserID,Password,PWOk);
      if PWOk then
      begin
        StrPCopy(szKey,UserId); StrCat(szKey,'\Time Constraints');
        StrPCopy(szTemp,WDays[WeekDay]); StrCat(szKey,szTemp);
        if RegDeleteKey(FAppKey,szKey) <> ERROR_SUCCESS then
          raise ESecurityBadDB.CreateFmt('Could not delete time constraint for user %s',[UserId])
        else
          Success := TRUE;
      end
      else
        raise ESecurityBadPW.Create('Invalid password for User ID');
    end
    else
      raise ESecurityBadDB.Create('Application database not found!');
  finally
    freemem(szKey,255);
    freemem(szTemp,255);
  end;
end;

procedure TThSecTime.SetWatchInterval(Value : integer);

begin
  if FWatchInterval <> Value then
  begin
    FWatchInterval := Value;
    FWatchTimer.Interval := Value;
    if FWatching then
      FWatchTimer.Enabled := TRUE
    else
      FWatchTimer.Enabled := FALSE;
  end;
end;

procedure TThSecTime.WatchTimer(Sender : TObject);

begin
  inc(MinuteCnt);
  if MinuteCnt = FWatchInterval then
  begin
    if assigned(FOnTimer) then FOnTimer(Sender);
    MinuteCnt := 0;
  end;
end;

end.
